using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace Smdn.Formats.Mime {
 public static class Decoder {
    public static MimeMessage Decode(MimeMessage message)
    {
      string decodedBody;
      byte[] decodedBodyBytes;
      Dictionary<string, string[]> decodedHeaders;

      DecodeBody(message, out decodedBody, out decodedBodyBytes);

      DecodeHeaders(message.Headers, out decodedHeaders);

      var decodedSubParts = new List<MimeMessage>();

      foreach (var part in message.SubParts) {
        decodedSubParts.Add(Decode(part));
      }

      var decoded = new MimeMessage(decodedHeaders, decodedBody, decodedSubParts);

      decoded.BodyBytes = decodedBodyBytes;

      return decoded;
    }

    public static string DecodeMimeEncodedString(string str)
    {
      if (str.Contains("=?") && str.Contains("?="))
        // デコードの必要あり?
        return encodedWordRegex.Replace(str, MimeEncodedStringMatchEvaluator);
      else
        // デコードの必要なし
        return str;
    }

    private static string MimeEncodedStringMatchEvaluator(Match match)
    {
      Encoding charset;

      // charset
      try {
        charset = Encoding.GetEncoding(match.Groups[1].Value);
      }
      catch {
        charset = latin1Encoding;
      }

      // encoding
      switch (match.Groups[2].Value.ToLower()) {
        case "b": // B-Encoding
          return charset.GetString(Convert.FromBase64String(match.Groups[3].Value));

        case "q": // Q-Encoding
          return charset.GetString(QuotedPrintable.FromQuotedPrintableString(match.Groups[3].Value));

        default: // unknown encoding
          return match.Value;
      }
    }

    private static void DecodeBody(MimeMessage message, out string body, out byte[] bodyBytes)
    {
      if (message.Headers.ContainsKey("Content-Transfer-Encoding")) {
        // Content-Transfer-Encodingに基づいてデコードする
        bodyBytes = DecodeByTransferEncoding(message.Headers["Content-Transfer-Encoding"][0], message.Body);
      }
      else {
        bodyBytes = Encoding.ASCII.GetBytes(message.Body);
      }

      var contentType = string.Empty;

      if (message.Headers.ContainsKey("Content-Type"))
        contentType = message.Headers["Content-Type"][0];

      if (contentType.StartsWith("text/")) {
        // Content-Type charsetに基づいてデコードする
        body = DecodeByContentTypeCharset(contentType, bodyBytes);
      }
      else if (contentType.StartsWith("multipart/")) {
        // multipartの場合は、text/plain; charset=US-ASCIIとして扱う
        body = Encoding.ASCII.GetString(bodyBytes);
      }
      else if (string.Empty.Equals(contentType)) {
        // Content-Typeが指定されていない(text/plain; charset=US-ASCIIとして扱う)
        body = Encoding.ASCII.GetString(bodyBytes);
      }
      else {
        // それ以外の場合は、バイナリと解釈する
        body = null;
      }
    }

    private static byte[] DecodeByTransferEncoding(string contentTransferEncoding, string text)
    {
      switch (contentTransferEncoding.ToLower()) {
        case "7bit":
          return Encoding.ASCII.GetBytes(text);

        case "8bit":
          return latin1Encoding.GetBytes(text);

        case "binary":
          return latin1Encoding.GetBytes(text);

        case "base64":
          return Convert.FromBase64String(text);

        case "quoted-printable":
          return QuotedPrintable.FromQuotedPrintableString(text);

        default:
          throw new NotSupportedException("unsupported transfer encoding");
      }
    }

    private static string DecodeByContentTypeCharset(string contentType, byte[] bytes)
    {
      var charset = GetCharset(contentType);

      if (charset == null)
        return latin1Encoding.GetString(bytes);
      else
        return charset.GetString(bytes);
    }

    private static Encoding GetCharset(string contentType)
    {
      var charset = Parser.GetHeaderParameter(contentType, "charset", true);

      if (string.Empty.Equals(charset))
        return null;

      try {
        return Encoding.GetEncoding(charset);
      }
      catch {
        return null;
      }
    }

    private static void DecodeHeaders(Dictionary<string, string[]> headers, out Dictionary<string, string[]> decodedHeaders)
    {
      decodedHeaders = new Dictionary<string,string[]>();

      foreach (var header in headers) {
        var decodedValue = new List<string>();
        foreach (var @value in header.Value) {
          decodedValue.Add(DecodeMimeEncodedString(@value));
        }
        decodedHeaders.Add(header.Key, decodedValue.ToArray());
      }
    }

    // encoded-word = "=?" charset "?" encoding "?" encoded-text "?="
    // charset = token    ; see section 3
    // encoding = token   ; see section 4
    // token = 1*<Any CHAR except SPACE, CTLs, and especials>
    // especials = "(" / ")" / "<" / ">" / "@" / "," / ";" / ":" / "
    //             <"> / "/" / "[" / "]" / "?" / "." / "="
    // encoded-text = 1*<Any printable ASCII character other than "?"
    //                   or SPACE>
    //               ; (but see "Use of encoded-words in message
    //               ; headers", section 5)
    private static readonly Regex encodedWordRegex = new Regex(@"\s*\=\?([^?]+)\?([^?]+)\?([^\?\s]+)\?\=\s*", RegexOptions.Singleline);
    private static readonly Encoding latin1Encoding = Encoding.GetEncoding("ISO-8859-1");
  }
}