using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4 {
  // 7.4.2. FETCH Response
  //       BODYSTRUCTURE
  //         The basic fields of a non-multipart body part are in the
  //         following order:
  //
  //         body type
  //            A string giving the content media type name as defined in
  //            [MIME-IMB].
  //
  //         body subtype
  //            A string giving the content subtype name as defined in
  //            [MIME-IMB].
  //
  //         body parameter parenthesized list
  //            A parenthesized list of attribute/value pairs [e.g., ("foo"
  //            "bar" "baz" "rag") where "bar" is the value of "foo" and
  //            "rag" is the value of "baz"] as defined in [MIME-IMB].
  //
  //         body id
  //            A string giving the content id as defined in [MIME-IMB].
  //
  //         body description
  //            A string giving the content description as defined in
  //            [MIME-IMB].
  //
  //         body encoding
  //            A string giving the content transfer encoding as defined in
  //            [MIME-IMB].
  //
  //         body size
  //            A number giving the size of the body in octets.  Note that
  //            this size is the size in its transfer encoding and not the
  //            resulting size after any decoding.
  //
  //         A body type of type MESSAGE and subtype RFC822 contains,
  //         immediately after the basic fields, the envelope structure,
  //         body structure, and size in text lines of the encapsulated
  //         message.
  //
  //         A body type of type TEXT contains, immediately after the basic
  //         fields, the size of the body in text lines.  Note that this
  //         size is the size in its content transfer encoding and not the
  //         resulting size after any decoding.
  public class ImapBodyStructure {
    public ImapBodyStructure[] Nested {
      get { return nested; }
    }

    /*
     * basic fields
     */
    public string Type {
      get { return type; }
    }

    public string Subtype {
      get { return subtype; }
    }

    public Dictionary<string, string> Parameters {
      get { return parameters; }
    }

    public string Id {
      get { return id; }
    }

    public string Description {
      get { return description; }
    }

    public string Encoding {
      get { return encoding; }
    }

    public int Size {
      get { return size; }
    }

    /*
     * type MESSAGE
     */
    public ImapEnvelope Envelope { // of encapsulated message
      get { return envelope; }
    }

    public ImapBodyStructure BodyStructure { // of encapsulated message
      get { return bodyStructure; }
    }

    /*
     * type TEXT, type MESSAGE
     */
    public int Lines { // the size of the body in text lines(type TEXT)
                        // size in text lines of the encapsulated message(type MESSAGE)
      get { return lines; }
    }

    public ImapBodyStructure(ImapBodyStructure[] nested,
                             string type,
                             string subtype,
                             Dictionary<string, string> parameters,
                             string id,
                             string description,
                             string encoding,
                             int size,
                             ImapEnvelope envelope,
                             ImapBodyStructure bodyStructure,
                             int lines)
    {
      this.nested         = nested ?? new ImapBodyStructure[] {};
      this.type           = type;
      this.subtype        = subtype;
      if (parameters == null)
        this.parameters   = new Dictionary<string, string>();
      else
        this.parameters   = new Dictionary<string, string>(parameters);
      this.id             = id;
      this.description    = description;
      this.encoding       = encoding;
      this.size           = size;
      this.envelope       = envelope;
      this.bodyStructure  = bodyStructure;
      this.lines          = lines;
    }

    public override string ToString()
    {
      return string.Format("{{Nested={0}, Type={1}, Subtype={2}, Parameters={3}, Id={4}, Description={5}, Encoding={6}, Size={7}, Envelope={8}, BodyStructure={9}, Lines={10}}}",
                           nested.Length,
                           type,
                           subtype,
                           parameters.Count,
                           id,
                           description,
                           encoding,
                           size,
                           envelope,
                           bodyStructure,
                           lines);
    }

    private readonly ImapBodyStructure[] nested;
    private readonly string type;
    private readonly string subtype;
    private readonly Dictionary<string, string> parameters;
    private readonly string id;
    private readonly string description;
    private readonly string encoding;
    private readonly int size;
    private readonly ImapEnvelope envelope;
    private readonly ImapBodyStructure bodyStructure;
    private readonly int lines;
  }
}
/*
body            = "(" (body-type-1part / body-type-mpart) ")"
body-extension  = nstring / number /
                   "(" body-extension *(SP body-extension) ")"
                    ; Future expansion.  Client implementations
                    ; MUST accept body-extension fields.  Server
                    ; implementations MUST NOT generate
                    ; body-extension fields except as defined by
                    ; future standard or standards-track
                    ; revisions of this specification.
body-ext-1part  = body-fld-md5 [SP body-fld-dsp [SP body-fld-lang
                  [SP body-fld-loc *(SP body-extension)]]]
                    ; MUST NOT be returned on non-extensible
                    ; "BODY" fetch
body-ext-mpart  = body-fld-param [SP body-fld-dsp [SP body-fld-lang
                  [SP body-fld-loc *(SP body-extension)]]]
                    ; MUST NOT be returned on non-extensible
                    ; "BODY" fetch
body-fields     = body-fld-param SP body-fld-id SP body-fld-desc SP
                  body-fld-enc SP body-fld-octets
body-fld-desc   = nstring
body-fld-dsp    = "(" string SP body-fld-param ")" / nil
body-fld-enc    = (DQUOTE ("7BIT" / "8BIT" / "BINARY" / "BASE64"/
                  "QUOTED-PRINTABLE") DQUOTE) / string
body-fld-id     = nstring
body-fld-lang   = nstring / "(" string *(SP string) ")"
body-fld-loc    = nstring
body-fld-lines  = number
body-fld-md5    = nstring
body-fld-octets = number
body-fld-param  = "(" string SP string *(SP string SP string) ")" / nil
body-type-1part = (body-type-basic / body-type-msg / body-type-text)
                  [SP body-ext-1part]
body-type-basic = media-basic SP body-fields
                    ; MESSAGE subtype MUST NOT be "RFC822"
body-type-mpart = 1*body SP media-subtype
                  [SP body-ext-mpart]
body-type-msg   = media-message SP body-fields SP envelope
                  SP body SP body-fld-lines
body-type-text  = media-text SP body-fields SP body-fld-lines

media-basic     = ((DQUOTE ("APPLICATION" / "AUDIO" / "IMAGE" /
                  "MESSAGE" / "VIDEO") DQUOTE) / string) SP
                  media-subtype
                    ; Defined in [MIME-IMT]
media-message   = DQUOTE "MESSAGE" DQUOTE SP DQUOTE "RFC822" DQUOTE
                    ; Defined in [MIME-IMT]
media-subtype   = string
                    ; Defined in [MIME-IMT]
media-text      = DQUOTE "TEXT" DQUOTE SP media-subtype
                    ; Defined in [MIME-IMT]
 */