using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4 {
  internal static class ImapConverter
  {
    internal static ImapAuthenticationType ToAuthType(string authType)
    {
      switch (authType.ToUpper()) {
        case "LOGIN":
          return ImapAuthenticationType.Login;
        case "PLAIN":
          return ImapAuthenticationType.Plain;
        case "CRAM-MD5":
          return ImapAuthenticationType.CramMd5;
        case "KERBEROS_V4":
          return ImapAuthenticationType.KerberosV4;
        case "*":
          return ImapAuthenticationType.Any;
        default:
          throw new NotSupportedException("unsupported authentication type");
      }
    }

    internal static string ToString(ImapAuthenticationType authType)
    {
      switch (authType) {
        case ImapAuthenticationType.Login:
          return "login";
        case ImapAuthenticationType.Plain:
          return "plain";
        case ImapAuthenticationType.CramMd5:
          return "cram-md5";
        case ImapAuthenticationType.KerberosV4:
          return "kerberos_v4";
        case ImapAuthenticationType.Any:
          return "*";
        case ImapAuthenticationType.None:
          return string.Empty;
        default:
          throw new ArgumentOutOfRangeException();
      }
    }

    internal static ImapListType ToListType(string listType)
    {
      switch (listType.ToUpper()) {
        case "LIST":
          return ImapListType.List;
        case "LSUB":
          return ImapListType.Lsub;
        default:
          throw new FormatException("invalid list type, it must be LIST or LSUB");
      }
    }

    internal static string ToString(ImapListType listType)
    {
      switch (listType) {
        case ImapListType.List:
          return "list";
        case ImapListType.Lsub:
          return "lsub";
        default:
          throw new ArgumentOutOfRangeException();
      }
    }

    internal static string ToString(ImapSystemFlags systemFlag)
    {
      return @"\" + systemFlag.ToString();
    }

    internal static string ToDateString(DateTime date)
    {
      return date.ToString("d-MMM-yyyy", CultureInfo.InvariantCulture);
    }

    internal static string ToDateTimeString(DateTime dateTime)
    {
      // date            = date-text / DQUOTE date-text DQUOTE
      // date-day        = 1*2DIGIT
      //                     ; Day of month
      // date-day-fixed  = (SP DIGIT) / 2DIGIT
      //                     ; Fixed-format version of date-day
      // date-month      = "Jan" / "Feb" / "Mar" / "Apr" / "May" / "Jun" /
      //                   "Jul" / "Aug" / "Sep" / "Oct" / "Nov" / "Dec"
      // date-text       = date-day "-" date-month "-" date-year
      // date-year       = 4DIGIT
      // date-time       = DQUOTE date-day-fixed "-" date-month "-" date-year
      //                   SP time SP zone DQUOTE
      // time            = 2DIGIT ":" 2DIGIT ":" 2DIGIT
      //                     ; Hours minutes seconds
      // zone            = ("+" / "-") 4DIGIT
      //                     ; Signed four-digit value of hhmm representing
      //                     ; hours and minutes east of Greenwich (that is,
      //                     ; the amount that the given time differs from
      //                     ; Universal Time).  Subtracting the timezone
      //                     ; from the given time will give the UT form.
      //                     ; The Universal Time zone is "+0000".
      return string.Format("\"{0} {1}\"",
                           dateTime.ToString("dd-MMM-yyyy HH:mm:ss", CultureInfo.InvariantCulture),
                           dateTime.ToString("zzz", CultureInfo.InvariantCulture).Replace(":", string.Empty));
    }

    // atom            = 1*ATOM-CHAR
    // ATOM-CHAR       = <any CHAR except atom-specials>
    // atom-specials   = "(" / ")" / "{" / SP / CTL / list-wildcards /
    //                   quoted-specials / resp-specials
    // flag-keyword    = atom
    // list-wildcards  = "%" / "*"
    // quoted-specials = DQUOTE / "\"
    // resp-specials   = "]"
    // CTL ::= <any ASCII control character and DEL,
    //          0x00 - 0x1f, 0x7f>
    internal static readonly char[] InvalidKeywordChars = new[] {
      '(', ')', '{', ' ', '%', '*', '"', '\\', ']',
      '\x00', '\x01', '\x02', '\x03', '\x04', '\x05', '\x06', '\x07', '\x08', '\x09', '\x0a', '\x0b', '\x0c', '\x0d', '\x0e', '\x0f',
      '\x10', '\x11', '\x12', '\x13', '\x14', '\x15', '\x16', '\x17', '\x18', '\x19', '\x1a', '\x1b', '\x1c', '\x1d', '\x1e', '\x1f',
      '\x7f',
    };

    // flag            = "\Answered" / "\Flagged" / "\Deleted" /
    //                   "\Seen" / "\Draft" / flag-keyword / flag-extension
    //                     ; Does not include "\Recent"
    internal static readonly ImapSystemFlags[] Flag = new[] {
      ImapSystemFlags.Seen,
      ImapSystemFlags.Answered,
      ImapSystemFlags.Flagged,
      ImapSystemFlags.Deleted,
      ImapSystemFlags.Draft,
    };

    internal static void ValidateKeywords(string[] keywords)
    {
      foreach (var keyword in keywords) {
        var index = keyword.IndexOfAny(InvalidKeywordChars);
        if (0 <= index)
          throw new ArgumentException(string.Format("'{0}' contains invalid charactor '{1}'", keyword, keyword[index]));
      }
    }

    internal static ImapParenthesizedString ToFlagList(ImapMessageFlag flags)
    {
      return ToFlagList(false, flags.SystemFlags, flags.Keywords);
    }

    // flag-list       = "(" [flag *(SP flag)] ")"
    internal static ImapParenthesizedString ToFlagList(bool validateKeywords, ImapSystemFlags systemFlags, params string[] keywords)
    {
      if (validateKeywords)
        ValidateKeywords(keywords);

      var flags = new List<ImapString>();

      foreach (var keyword in keywords) {
        flags.Add(keyword);
      }

      foreach (var systemFlag in Flag) {
        if ((systemFlags & systemFlag) != 0)
          flags.Add(ToString(systemFlag));
      }

      return new ImapParenthesizedString(flags.ToArray());
    }
  }
}