using System;
using System.Collections.Generic;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;

using Smdn.Protocols.Imap4.Session;

namespace Smdn.Protocols.Imap4.Client {
  // Layers:
  // * Client.ImapClient
  //   Client.ImapClientAgent
  //   Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  //   Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  public class ImapClient {
    protected ImapClientAgent Agent {
      get { return agent; }
    }

    public ImapSession Session {
      get { return session; }
      internal set
      {
        if (session != value) {
          session = value;
          if (session != null)
            session.Timeout = timeout;
        }
      }
    }

    public ImapUri HostUri {
      get
      {
        if (session == null)
          return hostUri;
        else
          return session.HostUri;
      }
      set
      {
        if (session == null) {
          if (value == null)
            throw new ArgumentNullException();
          else
            hostUri = value;
        }
        else {
          throw new InvalidOperationException("already logged in");
        }
      }
    }

    public string Password {
      get { return password; }
      set
      {
        if (session == null) {
          if (value == null)
            throw new ArgumentNullException();
          else
            password = value;
        }
        else {
          throw new InvalidOperationException("already logged in");
        }
      }
    }

    public bool UseSsl {
      get { return useSsl; }
      set
      {
        if (session == null)
          useSsl = value;
        else
          throw new InvalidOperationException("already logged in");
      }
    }

    public bool TryTls {
      get { return tryTls; }
      set
      {
        if (session == null)
          tryTls = value;
        else
          throw new InvalidOperationException("already logged in");
      }
    }

    public int Timeout {
      get { return timeout; }
      set
      {
        if (timeout != value) {
          timeout = value;
          if (session != null)
            session.Timeout = timeout;
        }
      }
    }

    public bool IsIdling {
      get
      {
        if (session == null)
          return false;
        else
          return session.IsIdling;
      }
    }

    public ImapMailbox SelectedMailbox {
      get
      {
        if (session == null)
          return null;
        else
          return session.SelectedMailbox;
      }
    }

    internal ImapClient(ImapClientAgent agent)
    {
      this.agent = agent;
    }

    /*
     * methods for authentication
     */

    /// <summary>tries to login</summary>
    public void Login()
    {
      agent.Connect(this);

      agent.Authenticate(this);
    }

    /// <summary>this method closes selected mailbox if required, and logs out</summary>
    public void Logout()
    {
      agent.Logout(this);
    }

    /*
     * methods for requesting by IMAP URL
     */

    /// <summary>fetches a message by IMAP URL</summary>
    /// <remarks>
    /// this method
    /// * creates session if client hasn't been connected,
    /// * authenticates if client hasn't been authenticated,
    /// </remarks>
    public ImapMessage FetchMessage(ImapUri messageUri)
    {
      return agent.FetchMessage(this, messageUri);
    }

    /// <summary>lists mailboxes by IMAP URL</summary>
    /// <remarks>
    /// this method
    /// * creates session if client hasn't been connected,
    /// * authenticates if client hasn't been authenticated,
    /// </remarks>
    public ImapMailbox[] ListMailboxes(ImapUri listMailboxUri)
    {
      return agent.ListMailboxes(this, listMailboxUri);
    }

    /// <summary>lists or searches messages by IMAP URL</summary>
    /// <remarks>
    /// this method
    /// * creates session if client hasn't been connected,
    /// * authenticates if client hasn't been authenticated,
    /// </remarks>
    public ImapMessage[] ListMessages(ImapUri listMessageUri)
    {
      return agent.ListMessages(this, listMessageUri);
    }

    /*
     * methods for requesting
     */

    /// <summary>fetches a message from listed message</summary>
    /// <remarks>
    /// this method
    /// * creates session if client hasn't been connected,
    /// * authenticates if client hasn't been authenticated,
    /// </remarks>
    public ImapMessage FetchMessage(ImapMessage message)
    {
      return FetchMessage(message.Uri);
    }

    /// <summary>lists messages in listed mailbox</summary>
    /// <remarks>
    /// this method
    /// * creates session if client hasn't been connected,
    /// * authenticates if client hasn't been authenticated,
    /// </remarks>
    public ImapMessage[] ListMessages(ImapMailbox mailbox)
    {
      return ListMessages(mailbox.Uri);
    }

    /// <summary>sends NOOP request</summary>
    public void NoOp()
    {
      agent.NoOp(this);
    }

    /// <summary>gets mailbox hierarchy</summary>
    public ImapMailboxHierarchy[] GetMailboxHierarchies()
    {
      var mailboxes = agent.GetExistMailboxes(this);

      return ImapMailboxHierarchy.Create(session.ServerInfo, mailboxes);
    }

    /// <summary>updates mailbox status</summary>
    public void UpdateMailboxStatus(ImapMailbox mailbox)
    {
      agent.UpdateMailboxStatus(this, mailbox);
    }

    /// <summary>begins idling</summary>
    public void BeginIdle()
    {
      agent.BeginIdle(this);
    }

    /// <summary>ends idling</summary>
    public void EndIdle()
    {
      agent.EndIdle(this);
    }

    private readonly ImapClientAgent agent;
    private ImapSession session = null;
    private ImapUri hostUri = null;
    private string password = string.Empty;
    private bool useSsl = false;
    private bool tryTls = true;
    private int timeout = 32 * 1000;
  }
}
