using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Security;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

using Smdn.Protocols.Imap4.Session;
using Smdn.Protocols.Imap4.Transaction;

namespace Smdn.Protocols.Imap4.Client {
  // Layers:
  //   Client.ImapClient
  // * Client.ImapClientAgent
  //   Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  //   Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  public class ImapClientAgent : IDisposable {
    public X509Certificate2Collection CertCollection {
      get { return agent.CertCollection; }
    }
    
    public RemoteCertificateValidationCallback ServerCertificateValidationCallback {
      get { return agent.ServerCertificateValidationCallback; }
      set { agent.ServerCertificateValidationCallback = value; }
    }

    public ImapClientAgent()
    {
    }

    public void Initialize()
    {
      if (agent != null)
        throw new InvalidOperationException("already initialized");

      Trace.Info("ImapClientAgent initializing");

      agent = new ImapSessionAgent();
      agent.CreateAuthenticatedStreamCallback = CreateAuthenticatedStream;
      agent.UpgradeAuthenticatedStreamCallback = UpgradeAuthenticatedStream;
      agent.Initialize();

      Trace.Info("ImapClientAgent initialized");
    }

    ~ImapClientAgent()
    {
      Dispose();
      GC.SuppressFinalize(this);
    }

    public void Dispose()
    {
      if (disposed)
        return;

      Trace.Info("ImapClientAgent disposing");

      if (agent != null) {
        ImapClient[] existClients;

        lock (clients) {
          existClients = clients.ToArray();
        }

        foreach (var client in existClients) {
          if (client.Session != null)
            Logout(client);
        }

        lock (clients) {
          clients.Clear();
        }

        agent.Dispose();
        agent = null;
      }

      Trace.Info("ImapClientAgent disposed");

      disposed = true;
    }

    public ImapClient CreateClient()
    {
      var client = new ImapClient(this);

      lock (clients) {
        clients.Add(client);
      }

      return client;
    }

    /*
     * methods for authentication and session management 
     */

    public void Connect(ImapClient client)
    {
      if (client.Session != null)
        throw new InvalidOperationException("already connected");

      if (client.HostUri == null)
        throw new InvalidOperationException("HostUri must be set");

      var session = agent.GetExistSession(client.HostUri);

      if (session != null)
        throw new InvalidOperationException("other client has already started requested session");

      if (!agent.Connect(client.HostUri, out session))
        throw new ImapClientExcetion("connect failed");

      client.Session = session;
    }

    /// <summary>sends CAPABILITY, AUTHENTICATE/LOGIN, NAMESPACE request</summary>
    public void Authenticate(ImapClient client)
    {
      try {
        var session = client.Session;

        if (session == null)
          throw new InvalidOperationException("not connected");

        for (var tryTls = 0; tryTls < 2; tryTls++) {
          if (session.ServerInfo.Capabilities.KnownCapabilities == ImapKnownCapabilities.None) {
            ImapCapability capability; // no use
            
            if (!agent.Capability(session, out capability)) {
              agent.Logout(session);
              client.Session = null;
              throw new ImapClientExcetion("capability comprehension failed");
            }
          }

          if (client.TryTls && tryTls == 0 &&
              session.ServerInfo.Capabilities.Has(ImapKnownCapabilities.StartTls))
            if (agent.StartTls(session))
              // conprehend capability again
              continue;

          break;
        }

        if (!agent.Authenticate(session, client.Password)) {
          if (!agent.Login(session, client.Password)) {
            agent.Logout(session);
            client.Session = null;
            throw new ImapClientExcetion("authentication was refused");
          }
        }

        ImapNamespace ns;

        // continue whether error occurred
        agent.Namespace(session, out ns);
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends CLOSE(if in selected state) and LOGOUT request</summary>
    public void Logout(ImapClient client)
    {
      RejectNonLoggedIn(client);

      try {
        if (client.Session.State == ImapSessionState.Selected)
          if (!agent.Close(client.Session))
            throw new ImapClientExcetion("closing mailbox failed");
      }
      finally {
        CheckConnectionAlive(client);
      }

      if (!agent.Logout(client.Session))
        throw new ImapClientExcetion("logout failed");

      client.Session = null;
    }

    /*
     * methods for requesting by IMAP URL
     */

    /// <summary>sends FETCH request by IMAP URL</summary>
    public ImapMessage FetchMessage(ImapClient client, ImapUri messageUri)
    {
      if (messageUri.Form != ImapUriForm.FetchMessage)
        throw new InvalidOperationException("the requested IMAP URL is invalid in this method");

      ImapResponse response;

      Request(client, messageUri, out response);
        
      if (!(response is ImapMessageResponse))
        throw new ImapClientExcetion("unexpected response received");

      if ((response as ImapMessageResponse).Messages.Length == 0)
        return null;
      else
        return (response as ImapMessageResponse).Messages[0];
    }

    /// <summary>sends LIST/LSUB request by IMAP URL</summary>
    public ImapMailbox[] ListMailboxes(ImapClient client, ImapUri listMailboxUri)
    {
      if (listMailboxUri.Form != ImapUriForm.ListMailboxes &&
          listMailboxUri.Form != ImapUriForm.Server)
        throw new InvalidOperationException("the requested IMAP URL is invalid in this method");

      ImapResponse response;

      Request(client, listMailboxUri, out response);

      if (!(response is ImapListResponse))
        throw new ImapClientExcetion("unexpected response received");

      return (response as ImapListResponse).Mailboxes;
    }

    /// <summary>sends SEARCH/FETCH request by IMAP URL</summary>
    public ImapMessage[] ListMessages(ImapClient client, ImapUri listMessageUri)
    {
      if (listMessageUri.Form != ImapUriForm.ListMessages &&
          listMessageUri.Form != ImapUriForm.SearchMessages)
        throw new InvalidOperationException("the requested IMAP URL is invalid in this method");

      ImapResponse response;

      Request(client, listMessageUri, out response);

      if (!(response is ImapMessageResponse))
        throw new ImapClientExcetion("unexpected response received");

      return (response as ImapMessageResponse).Messages;
    }

    /// <summary>sends request by IMAP URL</summary>
    /// <remarks>
    /// this method
    /// * list mailboxes if client hasn't LIST/LSUBed 
    /// * selects mailbox if no mailbox has been selected,
    /// </remarks>
    private void Request(ImapClient client, ImapUri requestUri, out ImapResponse response)
    {
      response = null;

      var ensureAuthenticated = false;
      var ensureSelected = false;
      
      switch (requestUri.Form) {
        case ImapUriForm.Server:
        case ImapUriForm.ListMailboxes:
          ensureAuthenticated = true;
          break;

        case ImapUriForm.ListMessages:
        case ImapUriForm.FetchMessage:
        case ImapUriForm.SearchMessages:
          ensureAuthenticated = true;
          ensureSelected = true;
          break;

        default:
          throw new UriFormatException("requested IMAP URL is invalid");
      }

      try {
        // not authenticated
        if (ensureAuthenticated && client.Session == null) {
          client.HostUri = new ImapUri(requestUri.GetLeftPart(UriPartial.Authority));

          Connect(client);

          Authenticate(client);
        }

        var session = client.Session;

        // not selected
        if (ensureSelected &&
            (session.SelectedMailbox == null ||
             session.SelectedMailbox != ImapMailboxManager.GetExistMailbox(session.ServerInfo, requestUri))
            ) {
          var mailboxName = requestUri.Mailbox;

          if (mailboxName.Contains("/")) {
            // replace delimiter
            var rootMailboxName = mailboxName.Substring(0, mailboxName.IndexOf("/"));
            var rootMailbox = ImapMailboxManager.GetExistMailbox(session.ServerInfo, rootMailboxName);

            if (rootMailbox == null || string.Empty.Equals(rootMailbox.HierarchyDelimiter)) {
              ImapMailbox[] mailboxes;
              if (!agent.List(session, rootMailboxName, out mailboxes) || mailboxes.Length == 0)
                throw new ImapClientExcetion("replacing mailbox hierarchy delimiter failed");

              rootMailbox = mailboxes[0];
            }

            mailboxName = mailboxName.Replace("/", rootMailbox.HierarchyDelimiter);
          }

          if (!agent.Select(session, mailboxName))
            throw new ImapClientExcetion("selecting mailbox failed");
        }

        // request
        if (!agent.Request(session, requestUri, out response))
          throw new ImapClientExcetion("request failed");
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    private ImapMailbox[] ListLsubMailboxes(ImapSession session)
    {
      var allMailboxes = new List<ImapMailbox>();

      ImapMailbox[] mbx;

      if (!agent.List(session,"INBOX", out mbx))
        throw new ImapClientExcetion("listing INBOX failed");

      allMailboxes.AddRange(mbx);

      if (!agent.Lsub(session, string.Empty, out mbx))
        throw new ImapClientExcetion("listing mailbox failed");

      allMailboxes.AddRange(mbx);

      return allMailboxes.ToArray();
    }

    /*
     * methods for requesting 
     */

    /// <summary>sends NOOP request</summary>
    public void NoOp(ImapClient client)
    {
      RejectNonLoggedIn(client);

      try {
        // continue whether error occurred
        agent.NoOp(client.Session);
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>gets exist mailboxes</summary>
    /// <remarks>
    /// this method list mailboxes if client hasn't LIST/LSUBed
    /// </remarks>
    public ImapMailbox[] GetExistMailboxes(ImapClient client)
    {
      RejectNonLoggedIn(client);

      var mailboxes = ImapMailboxManager.GetExistMailboxes(client.Session.ServerInfo);

      if (mailboxes != null)
        return mailboxes;

      try {
        return ListLsubMailboxes(client.Session);
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /// <summary>sends STATUS request and update mailbox status</summary>
    public void UpdateMailboxStatus(ImapClient client, ImapMailbox mailbox)
    {
      RejectNonLoggedIn(client);

      try {
        if (!agent.Status(client.Session, mailbox, ImapStatusDataItem.All))
          throw new ImapClientExcetion("updating mailbox status failed");
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    public void BeginIdle(ImapClient client)
    {
      RejectNonLoggedIn(client);

      try {
        if (!agent.BeginIdle(client.Session, null))
          throw new ImapClientExcetion("idling failed");
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    public void EndIdle(ImapClient client)
    {
      RejectNonLoggedIn(client);

      try {
        if (!agent.EndIdle(client.Session))
          throw new ImapClientExcetion("idle completion failed");
      }
      finally {
        CheckConnectionAlive(client);
      }
    }

    /*
     * private methods 
     */

    private void RejectNonLoggedIn(ImapClient client)
    {
      if (client.Session == null)
        throw new InvalidOperationException("not logged in");
    }

    private void CheckConnectionAlive(ImapClient client)
    {
      if (client.Session == null)
        return;

      if (client.Session.State == ImapSessionState.NotConnected) {
        client.Session = null;

        throw new ImapDisconnectedExcetion("session was disconnected by server");
      }
    }

    private ImapClient GetExistClientFromUri(ImapUri uri)
    {
      ImapClient[] existClients;

      lock (clients) {
        existClients = clients.ToArray();
      }

      foreach (var client in existClients) {
        if (client.HostUri.UserAtHost == uri.UserAtHost)
          return client;
      }

      return null;
    }

    private Stream CreateAuthenticatedStream(ImapUri uri, Stream baseStream)
    {
      var client = GetExistClientFromUri(uri);

      if (client != null && client.UseSsl)
        return UpgradeAuthenticatedStream(uri, baseStream);
      else
        return baseStream;
    }

    private Stream UpgradeAuthenticatedStream(ImapUri uri, Stream baseStream)
    {
      SslStream sslStream = null;

      if (agent.ServerCertificateValidationCallback == null)
        sslStream = new SslStream(baseStream, false);
      else
        sslStream = new SslStream(baseStream, false, agent.ServerCertificateValidationCallback);

      sslStream.AuthenticateAsClient(uri.Host, agent.CertCollection, SslProtocols.Default, true);

      Trace.Info("created SSL stream for {0}", uri);

      return sslStream;
    }

    private bool disposed = false;
    private ImapSessionAgent agent = null;
    private readonly List<ImapClient> clients = new List<ImapClient>();
  }
}
