using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Threading;

namespace Smdn.Protocols.Imap4.Protocol {
  // Layers:
  //   Client.ImapClient
  //   Client.ImapClientAgent
  //   Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  // * Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  public class ImapConnection {
    private class Receiver : IDisposable {
      public Receiver(Stream stream)
      {
        this.stream = stream;
      }

      public void Dispose()
      {
        readEvent.Close();
      }

      public byte[] TryReceive()
      {
        lock (receivedData) {
          if (0 < receivedData.Count)
            return receivedData.Dequeue();
        }

        if (readEvent.WaitOne(0, false))
          BeginReceive();

        return null;
      }

      private void BeginReceive()
      {
        stream.BeginRead(buffer, 0, buffer.Length, EndReceive, this);
      }

      private void EndReceive(IAsyncResult ar)
      {
        // TODO: サーバからの切断の考慮
        var read = stream.EndRead(ar);

        if (read <= 0)
          return;

        var data = new byte[read];

        Array.Copy(buffer, data, read);

        lock (receivedData) {
          receivedData.Enqueue(data);
        }

        Trace.Dump(data);

        readEvent.Set();
      }

      private readonly Stream stream;
      private readonly AutoResetEvent readEvent = new AutoResetEvent(true);
      private readonly byte[] buffer = new byte[0x400];
      private readonly Queue<byte[]> receivedData = new Queue<byte[]>();
    }

    /*
    private class Sender {
      // TODO: impl
      public Sender(Stream stream)
      {
        this.stream = stream;
      }

      private readonly Stream stream;
    }
    */

    public abstract class DividedCommandContext {
    }

    private class DividedCommandContextImpl : DividedCommandContext {
      public byte[] Command {
        get { return command; }
      }

      public DividedCommandContextImpl(byte[] command, int[] division)
      {
        this.command = command;
        this.division = division;
      }

      public void GetNextRange(out int start, out int count, out int remain)
      {
        if (index == 0) {
          start = 0;
          if (division.Length == 0) {
            count = command.Length;
            remain = 0;
            return;
          }
          else {
            count = division[index];
          }
        }
        else {
          start = division[index - 1];
          if (index == division.Length)
            count = command.Length - start;
          else
            count = division[index] - start;
        }

        remain = command.Length - (count + start);
        index++;
      }

      private readonly byte[] command;
      private readonly int[] division;
      private int index = 0;
    }

    /*
     * properties
     */

    internal int Id {
      get { return id; }
    }

    public bool Connected {
      get { return connected; }
    }

    public bool IsIdling {
      get { return isIdling; }
      internal set { isIdling = value; }
    }

    internal ImapSizeStatus UpdatedStatusWhileIdling {
      get { return updatedStatusWhileIdling; }
      set { updatedStatusWhileIdling = value; }
    }

    public ImapSizeStatusUpdatedCallback IdlingStatusUpdatedCallback {
      get { return idlingStatusUpdatedCallback; }
      internal set { idlingStatusUpdatedCallback = value; }
    }

    public ImapServerInfo ServerInfo {
      get { return serverInfo; }
    }

    // for '5.4. Autologout Timer'
    public DateTime LastSentTime {
      get { return lastSentTime; }
    }

    public int QueuedResponseCount {
      get { return parsedResponse.Count; }
    }

    private ImapConnection(ImapUri serverUri, int id)
    {
      this.serverInfo = new ImapServerInfo(serverUri);
      this.id = id;
    }

    ~ImapConnection()
    {
      Disconnect();
    }

    static public ImapConnection Connect(ImapUri serverUri, ImapCreateAuthenticatedStreamCallback createAuthStreamCallback)
    {
      var connection = new ImapConnection(serverUri, Interlocked.Increment(ref connectionIdCount));

      if (connectionIdCount == int.MaxValue)
        Interlocked.Exchange(ref connectionIdCount, 0);

      connection.client = new TcpClient();

      if (serverUri.Port == -1)
        connection.client.Connect(serverUri.Host, ImapUri.DefaultPort);
      else
        connection.client.Connect(serverUri.Host, serverUri.Port);

      connection.stream = connection.client.GetStream();

      if (createAuthStreamCallback == null)
        connection.CreateSenderAndReceiver();
      else
        connection.UpgradeSecurity(serverUri, createAuthStreamCallback);

      connection.connected = true;

      Trace.Info("CID:{0} connected to {1} ({2})", connection.Id, connection.client.Client.RemoteEndPoint, serverUri.Host);

      return connection;
    }

    public void Disconnect()
    {
      if (!connected)
        return;

      Trace.Info("CID:{0} disconnecting from {1}", id, client.Client.RemoteEndPoint);

      if (receiver != null) {
        receiver.Dispose();
        receiver = null;
      }

      // if (sender != null) {
      //   sender.Dispose();
      //   sender = null;
      // }

      if (client != null) {
        client.Close();
        client = null;
      }

      Trace.Info("CID:{0} disconnected", id);

      connected = false;
    }

    public ImapProtocolCommand CreateCommand(string command, params ImapString[] arguments)
    {
      var comm = new ImapProtocolCommand(command, commandTag.ToString("D4"), arguments);

      if (Interlocked.Increment(ref commandTag) == 0x10000)
        Interlocked.Exchange(ref commandTag, 0);

      return comm;
    }

    public ImapProtocolCommand CreateContinuingCommand(params ImapString[] arguments)
    {
      return new ImapProtocolCommand(null, null, arguments);
    }

    public DividedCommandContext SendCommand(ImapProtocolCommand command)
    {
      int[] division;

      return SendCommand(new DividedCommandContextImpl(command.Format(out division), division));
    }

    public DividedCommandContext SendCommand(DividedCommandContext context)
    {
      var c = (context as DividedCommandContextImpl);

      int start, count, remain;

      c.GetNextRange(out start, out count, out remain);

      Trace.Log(this, c.Command, start, count);
      Trace.Dump(c.Command, start, count);

      stream.Write(c.Command, start, count);

      lastSentTime = DateTime.Now;

      if (remain == 0)
        return null;
      else
        return c;
    }

    public ImapProtocolResponse TryReceiveResponse()
    {
      // return parsed response
      if (0 < parsedResponse.Count)
        return parsedResponse.Dequeue();

      // get received data
      byte[] received = receiver.TryReceive();

      if (received == null)
        return null;

      // parse received data
      foreach (var response in parser.Parse(received)) {
        Trace.Log(this, response);
        Trace.Dump(response);

        parsedResponse.Enqueue(response);
      }

      // return parsed response
      if (0 < parsedResponse.Count)
        return parsedResponse.Dequeue();

      return null;
    }

    public override string ToString()
    {
      return string.Format("{{Host={0}, Connected={1}, Id={2}}}",
                           client.Client.RemoteEndPoint,
                           client.Client.Connected,
                           id);
    }

    internal void UpgradeSecurity(ImapUri serverUri, ImapCreateAuthenticatedStreamCallback createAuthStreamCallback)
    {
      stream = createAuthStreamCallback(serverUri, stream);

      CreateSenderAndReceiver();
    }

    private void CreateSenderAndReceiver()
    {
      if (receiver != null)
        receiver.Dispose();

      receiver = new Receiver(stream);

      // for first receiving
      receiver.TryReceive();

      //if (sender != null)
      //  sender.Dispose();

      //sender = new Sender(stream);
    }

    private static int connectionIdCount = 0;
    private readonly int id = 0;
    private bool connected = false;
    private bool isIdling = false;
    private ImapSizeStatus updatedStatusWhileIdling = null;
    private readonly ImapServerInfo serverInfo;
    private DateTime lastSentTime = DateTime.Now;
    private TcpClient client;
    private Stream stream;
    private int commandTag = 0;
    private ImapSizeStatusUpdatedCallback idlingStatusUpdatedCallback = null;
    private readonly ImapProtocolResponseParser parser = new ImapProtocolResponseParser();
    private readonly Queue<ImapProtocolResponse> parsedResponse = new Queue<ImapProtocolResponse>();
    private Receiver receiver = null;
    //private Sender sender = null;
  }
}
