using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4.Protocol {
  // format converting methods for ImapDataResponse
  public static class ImapDataResponseConverter {
    /*
     * 6. Client Commands
     */

    // 6.1.2. NOOP Command
    //    Responses:  no specific responses for this command (but see below)
    //       Since any command can return a status update as untagged data, the
    //       NOOP command can be used as a periodic poll for new messages or
    //       message status updates during a period of inactivity (this is the
    //       preferred method to do this).

    //    Example:    C: a002 NOOP
    //                S: a002 OK NOOP completed
    //                   . . .
    //                C: a047 NOOP
    //                S: * 22 EXPUNGE
    //                S: * 23 EXISTS
    //                S: * 3 RECENT
    //                S: * 14 FETCH (FLAGS (\Seen \Deleted))
    //                S: a047 OK NOOP completed

    // 6.4.1. CHECK Command
    //    Responses:  no specific responses for this command
    //       There is no guarantee that an EXISTS untagged response will happen
    //       as a result of CHECK.  NOOP, not CHECK, SHOULD be used for new
    //       message polling.
    public static bool FromNoOpCheckCommandResponse(ImapDataResponse response, out ImapMessage messageData, out int? expunge, out int? exists, out int? recent)
    {
      return FromSizeStatusResponse(response, out messageData, out expunge, out exists, out recent);
    }

    // 6.3.1. SELECT Command
    //   Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
    //               REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
    //               UIDNEXT, UIDVALIDITY
    // 6.3.2. EXAMINE Command
    //   Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
    //               REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
    //               UIDNEXT, UIDVALIDITY
    public static bool FromSelectExamineCommandResponse(ImapDataResponse response, out ImapMessageFlag flags, out int? exists, out int? recent)
    {
      flags         = null;
      exists        = null;
      recent        = null;

      // FLAGS
      ImapMessageFlag respFlags;

      if (FromFlagsResponse(response, out respFlags)) {
        flags = respFlags;
        return true;
      }

      // EXISTS
      int respExists;

      if (FromExistsResponse(response, out respExists)) {
        exists = respExists;
        return true;
      }

      // RECENT
      int respRecent;

      if (FromRecentResponse(response, out respRecent)) {
        recent = respRecent;
        return true;
      }

      return false;
    }

    // 6.3.11. APPEND Command
    //    Responses:  no specific responses for this command
    //       If the mailbox is currently selected, the normal new message
    //       actions SHOULD occur.  Specifically, the server SHOULD notify the
    //       client immediately via an untagged EXISTS response.  If the server
    //       does not do so, the client MAY issue a NOOP command (or failing
    //       that, a CHECK command) after one or more APPEND commands.
    public static bool FromAppendCommandResponse(ImapDataResponse response, out ImapMessage messageData, out int? expunge, out int? exists, out int? recent)
    {
      return FromSizeStatusResponse(response, out messageData, out expunge, out exists, out recent);
    }

    // 6.4.5. FETCH Command
    //    Responses:  untagged responses: FETCH
    public static bool FromFetchCommandResponse(ImapDataResponse response, out ImapMessage messageData)
    {
      return FromFetchResponse(response, out messageData);
    }

    // 6.4.6. STORE Command
    //    Responses:  untagged responses: FETCH
    public static bool FromStoreCommandResponse(ImapDataResponse response, out ImapMessage messageData)
    {
      return FromFetchResponse(response, out messageData);
    }

    // EXISTS, EXPUNGE, RECENT, RECENT
    private static bool FromSizeStatusResponse(ImapDataResponse response, out ImapMessage messageData, out int? expunge, out int? exists, out int? recent)
    {
      messageData = null;
      expunge = null;
      exists = null;
      recent = null;

      // EXISTS
      int respExists;

      if (FromExistsResponse(response, out respExists)) {
        exists = respExists;
        return true;
      }

      // EXPUNGE
      int respExpunge;

      if (FromExpungeResponse(response, out respExpunge)) {
        expunge = respExpunge;
        return true;
      }

      // RECENT
      int respRecent;

      if (FromRecentResponse(response, out respRecent)) {
        recent = respRecent;
        return true;
      }

      // FETCH
      if (FromFetchResponse(response, out messageData))
        return true;

      return false;
    }

    /*
     * 7.2. Server Responses - Server and Mailbox Status
     */

    // 7.2.1 CAPABILITY Response
    //   Contents:   capability listing
    //   Example:    S: * CAPABILITY IMAP4rev1 STARTTLS AUTH=GSSAPI XPIG-LATIN
    public static bool FromCapabilityResponse(ImapDataResponse response, out ImapCapability capability)
    {
      // 1. check response format
      // 2. parse response
      // 3. set result
      return ImapDataConverter.ToCapabilityData(response.Data, out capability);
    }

    // 7.2.2 LIST Response
    //   Contents:   name attributes
    //               hierarchy delimiter
    //               name
    //   Example:    S: * LIST (\Noselect) "/" ~/Mail/foo
    // 7.2.3 LSUB Response
    //   Contents:   name attributes
    //               hierarchy delimiter
    //               name
    //   Example:    S: * LSUB () "." #news.comp.mail.misc
    public static bool FromListLsubResponse(ImapDataResponse response, out ImapMailboxFlag nameAttributes, out string hierarchyDelimiter, out string name)
    {
      nameAttributes = null;
      hierarchyDelimiter = null;
      name = string.Empty;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text},
      }, ImapDataFormat.None))
        return false;

      if (!(ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "LIST") ||
            ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "LSUB")))
        return false;

      // 2. parse response
      // 3. set result
      hierarchyDelimiter = ImapDataConverter.ToNString(response.Data[2]);
      name = ImapEncoding.FromModifiedUTF7StringNoException(response.Data[3].Text);

      return ImapDataConverter.ToMbxListFlags(response.Data[1].List, out nameAttributes);
    }

    // 7.2.4 STATUS Response
    //   Contents:   name
    //               status parenthesized list
    //   Example:    S: * STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292)
    public static bool FromStatusResponse(ImapDataResponse response, out string name, out int? messages, out int? recent, out int? unseen, out int? uidNext, out int? uidValidity)
    {
      /*
      status          = "STATUS" SP mailbox SP
                        "(" status-att *(SP status-att) ")"
      */

      name        = string.Empty;
      messages    = null;
      recent      = null;
      unseen      = null;
      uidNext     = null;
      uidValidity = null;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
      }, ImapDataFormat.Text))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "STATUS"))
        return false;

      // 2. parse response
      // 3. set result
      name = ImapEncoding.FromModifiedUTF7StringNoException(response.Data[1].Text);

      return ImapDataConverter.ToStatusAtt(response.Data[2].List, out messages, out recent, out unseen, out uidNext, out uidValidity);
    }

    // 7.2.5 SEARCH Response
    //   Contents:   zero or more numbers
    //   Example:    S: * SEARCH 2 3 6
    public static bool FromSearchResponse(ImapDataResponse response, out int[] messages)
    {
      messages = new int[] {};

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
      }, ImapDataFormat.Text))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "SEARCH"))
        return false;

      // 2. parse response
      var msgs = new List<int>();

      for (var index = 1; index < response.Data.Length; index++) {
        int? msg = ImapDataConverter.ToNumber(response.Data[index]);

        if (msg != null)
          msgs.Add((int)msg);
      }

      // 3. set result
      messages = msgs.ToArray();

      return true;
    }

    // 7.2.6 FLAGS Response
    //   Contents:   flag parenthesized list
    //   Example:    S: * FLAGS (\Answered \Flagged \Deleted \Seen \Draft)
    public static bool FromFlagsResponse(ImapDataResponse response, out ImapMessageFlag flags)
    {
      flags = null;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
      }, ImapDataFormat.None))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "FLAGS"))
        return false;

      // 2. parse response
      // 3. set result
      return ImapDataConverter.ToFlag(response.Data[1].List, out flags);
    }

    /*
     * 7.3. Server Responses - Mailbox Size
     */

    // 7.3.1. EXISTS Response
    //   Contents:   none
    //   Example:    S: * 23 EXISTS
    public static bool FromExistsResponse(ImapDataResponse response, out int exists)
    {
      return FromSizeOrSequenceNumberResponse(response, "EXISTS", out exists);
    }

    // 7.3.2. RECENT Response
    //   Contents:   none
    //   Example:    S: * 5 RECENT
    public static bool FromRecentResponse(ImapDataResponse response, out int recent)
    {
      return FromSizeOrSequenceNumberResponse(response, "RECENT", out recent);
    }

    /*
     * 7.4. Server Responses - Message Status
     */

    // 7.4.1. EXPUNGE Response
    //   Contents:   none
    //   Example:    S: * 44 EXPUNGE
    public static bool FromExpungeResponse(ImapDataResponse response, out int expunge)
    {
      return FromSizeOrSequenceNumberResponse(response, "EXPUNGE", out expunge);
    }

    // 7.4.2. FETCH Response
    //   Contents:   message data
    //   Example:    S: * 23 FETCH (FLAGS (\Seen) RFC822.SIZE 44827)
    public static bool FromFetchResponse(ImapDataResponse response, out ImapMessage messageData)
    {
      /*
      message-data    = nz-number SP ("EXPUNGE" / ("FETCH" SP msg-att))
       */

      messageData = null;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
      }, ImapDataFormat.None))
        return false;

      // get sequence number
      int sequence;

      if (!FromSizeOrSequenceNumberResponse(response, "FETCH", out sequence))
        return false;

      // 2. parse response
      ImapMessageFlag       flags           = null;
      ImapEnvelope          envelope        = null;
      DateTime?             internalDate    = null;
      // TODO: rfc822header = null;
      // TODO: rfc822text = null;
      int?                  size            = null;
      ImapBodyStructure     bodyStructure   = null;
      string               body            = null;
      int?                  uid             = null;

      if (!ImapDataConverter.ToMsgAtt(response.Data[2].List,
                    out flags,
                    out envelope,
                    out internalDate,
                    out size,
                    out bodyStructure,
                    out body,
                    out uid))
        return false;

      // 3. set result
      messageData = new ImapMessage(sequence);

      if (flags != null)          messageData.Flags = flags;
      if (envelope != null)       messageData.Envelope = envelope;
      if (internalDate != null)   messageData.InternalDate = (DateTime)internalDate;
      if (size != null)           messageData.BodySize = (int)size;
      if (bodyStructure != null)  messageData.BodyStructure = bodyStructure;
      if (body != null)           messageData.Body = body;
      if (uid != null)            messageData.Uid = (int)uid;

      return true;
    }

    //   Example:    S: * 23 EXISTS
    //   Example:    S: * 5 RECENT
    //   Example:    S: * 44 EXPUNGE
    //   Example:    S: * 23 FETCH (FLAGS (\Seen) RFC822.SIZE 44827)
    private static bool FromSizeOrSequenceNumberResponse(ImapDataResponse response, string responseType, out int number)
    {
      number = 0;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
      }, ImapDataFormat.None))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[1], responseType))
        return false;

      // 2. parse response
      // 3. set result
      if (!int.TryParse(response.Data[0].Text, out number))
        return false;

      return true;
    }

    /*
     * RFC 2177 IMAP4 IDLE command
     * http://tools.ietf.org/html/rfc2177
     */

    // 3. Specification
    //   IDLE Command
    //   The IDLE command remains active until the client
    //   responds to the continuation, and as long as an IDLE command is
    //   active, the server is now free to send untagged EXISTS, EXPUNGE, and
    //   other messages at any time
    public static bool FromIdleCommandResponse(ImapDataResponse response, out ImapMessage messageData, out int? expunge, out int? exists, out int? recent)
    {
      return FromSizeStatusResponse(response, out messageData, out expunge, out exists, out recent);
    }

    /*
     * RFC 2342 IMAP4 Namespace
     * http://tools.ietf.org/html/rfc2342
     */

    // 5. NAMESPACE Command
    //   Response:  an untagged NAMESPACE response that contains the prefix
    //                and hierarchy delimiter to the server's Personal
    //                Namespace(s), Other Users' Namespace(s), and Shared
    //                Namespace(s) that the server wishes to expose. The
    //                response will contain a NIL for any namespace class
    //                that is not available. Namespace_Response_Extensions
    //                MAY be included in the response.
    //                Namespace_Response_Extensions which are not on the IETF
    //                standards track, MUST be prefixed with an "X-".
    //   Example 5.4:
    //   S: * NAMESPACE (("" "/")) (("~" "/")) (("#shared/" "/")
    //      ("#public/" "/")("#ftp/" "/")("#news." "."))
    public static bool FromNamespaceResponse(ImapDataResponse response, out ImapNamespace ns)
    {
      ns = null;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List, ImapDataFormat.Nil},
        new [] {ImapDataFormat.List, ImapDataFormat.Nil},
        new [] {ImapDataFormat.List, ImapDataFormat.Nil},
      }, ImapDataFormat.None))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "NAMESPACE"))
        return false;

      // 2. parse response
      var nsgroups = new ImapNamespace.NamespaceGroup[3];

      for (int index = 0; index < 3; index++) {
        if (response.Data[index + 1].Format == ImapDataFormat.Nil) {
          // TODO: 名前空間がNILな場合の考慮
          nsgroups[index] = new ImapNamespace.NamespaceGroup(string.Empty, string.Empty, string.Empty);
        }
        else {
          foreach (var nsdata in response.Data[index + 1].List) {
            // TODO: 複数の名前空間が定義されている場合の考慮
            if (3 == nsdata.List.Length) {
              nsgroups[index] = new ImapNamespace.NamespaceGroup(ImapEncoding.FromModifiedUTF7StringNoException(nsdata.List[0].Text),
                                                                 nsdata.List[1].Text,
                                                                 ImapEncoding.FromModifiedUTF7StringNoException(nsdata.List[2].Text)); // TODO: Namespace_Response_Extension = SP string SP "(" string *(SP string) ")"
            }
            else if (2 == nsdata.List.Length) {
              nsgroups[index] = new ImapNamespace.NamespaceGroup(ImapEncoding.FromModifiedUTF7StringNoException(nsdata.List[0].Text),
                                                                 nsdata.List[1].Text,
                                                                 string.Empty);
            }
            else {
              return false;
            }
          }
        }
      }

      // 3. set result
      ns = new ImapNamespace(nsgroups[0], nsgroups[1], nsgroups[2]);

      return true;
    }
  }
}
