using System;
using System.Collections.Generic;
using System.Text;

namespace Smdn.Protocols.Imap4.Protocol {
  // Layers:
  //   Client.ImapClient
  //   Client.ImapClientAgent
  //   Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  // * Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  // handles '6. Client Commands'
  public class ImapProtocolCommand {
    public string Tag {
      get { return tag; }
    }

    public string Command {
      get { return command; }
    }

    public ImapString[] Arguments {
      get { return arguments; }
    }

    public ImapProtocolCommand(string command)
      : this(command, string.Empty, new ImapString[] {})
    {
    }

    public ImapProtocolCommand(string command, string tag, params ImapString[] arguments)
    {
      this.command = command;
      this.tag = tag;
      this.arguments = arguments;
    }

    public override string ToString()
    {
      return string.Format("{{Command={0}, Tag={1}}}", command, tag);
    }

    private List<byte> Format(IEnumerable<ImapString> args, out int[] divisionOctets)
    {
      var formatted = new List<byte>();
      var division = new List<int>();

      foreach (var arg in args) {
        if (formatted.Count != 0)
          formatted.Add(0x20); // SP

        if (arg is ImapStringList) {
          var parenthesize = (arg is ImapParenthesizedString);

          if (parenthesize)
            formatted.Add(0x28); // '('

          var offset = formatted.Count;
          int[] div;

          formatted.AddRange(Format(arg as ImapStringList, out div));

          foreach (var d in div) {
            division.Add(d + offset);
          }

          if (parenthesize)
            formatted.Add(0x29); // ')'
        }
        else {
          if (arg is ImapLiteralString) {
            var literal = arg as ImapLiteralString;

            foreach (var c in string.Format("{{{0}}}", literal.Octets)) {
              formatted.Add((byte)c);
            }

            formatted.Add(0x0d); // CR
            formatted.Add(0x0a); // LF

            // divide to '{number} CRLF' and '*CHAR8'
            division.Add(formatted.Count);

            formatted.AddRange(literal.Charset.GetBytes(literal.Value));
          }
          else if (arg is ImapQuotedString) {
            formatted.Add(0x22); // DQUOTE
            formatted.AddRange(ImapEncoding.Encoding7Bit.GetBytes(arg.Value));
            formatted.Add(0x22); // DQUOTE
          }
          else {
            formatted.AddRange(ImapEncoding.Encoding7Bit.GetBytes(arg.Value));
          }
        }
      }

      divisionOctets = division.ToArray();

      return formatted;
    }

    public byte[] Format(out int[] divisionOctets)
    {
      var args = new List<ImapString>();

      if (command != null) {
        args.Add(new ImapString(tag));
        args.Add(new ImapString(command));
      }

      args.AddRange(arguments);

      var formatted = Format(args, out divisionOctets);

      formatted.Add(0x0d); // CR
      formatted.Add(0x0a); // LF

      return formatted.ToArray();
    }

    private readonly string command;
    private readonly string tag;
    private readonly ImapString[] arguments;
  }
}