using System;
using System.Collections.Generic;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal class AppendTransaction : TransactionBase {

    // ProcessArgumentNotSetted
    //   x Conclude
    // ProcessAppend
    //   ->WaitForAppend
    // WaitForAppend
    //   x Conclude

    public AppendTransaction(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("mailbox name"))
        return ProcessArgumentNotSetted;
      else if (!Request.Arguments.ContainsKey("message literal"))
        return ProcessArgumentNotSetted;
      else
        return ProcessAppend;
    }

    private void ProcessArgumentNotSetted()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "arguments 'mailbox name' and 'message literal' must be setted"));
    }

    private void ProcessAppend()
    {
      var arguments = new List<ImapString>();

      arguments.Add(new ImapQuotedString(ImapEncoding.ToModifiedUTF7String((string)Request.Arguments["mailbox name"])));

      if (Request.Arguments.ContainsKey("flag parenthesized list"))
        arguments.Add(Request.Arguments["flag parenthesized list"]);

      if (Request.Arguments.ContainsKey("date/time string"))
        arguments.Add(Request.Arguments["date/time string"]);

      arguments.Add(Request.Arguments["message literal"]);

      Send("append", WaitForAppend, arguments.ToArray());
    }

    private void WaitForAppend()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapCommandContinuationRequest) {
        // send literals
        if (!SendContinue())
          Conclude(new ImapResponse(ImapTransactionStatus.Forbidden, "sending literals failed"));
      }
      else if (response is ImapDataResponse) {
        ImapMessage messageData;
        int? expunge;
        int? exists;
        int? recent;
        if (ImapDataResponseConverter.FromAppendCommandResponse(response as ImapDataResponse,
                                                                out messageData,
                                                                out expunge,
                                                                out exists,
                                                                out recent)) {
          if (messageData != null)    updatedMessages.Add(messageData);
          if (expunge != null)        updatedExpunge.Add((int)expunge);
          if (exists != null)         updatedExists = exists;
          if (recent != null)         updatedRecent = recent;
        }
      }
      else if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.Tag != LastTag)
          return;

        if (status.Status == ImapProtocolStatus.Ok) {
          Conclude(new ImapSizeStatusUpdatedResponse(ImapTransactionStatus.Ok,
                                                new ImapSizeStatus(updatedMessages.ToArray(),
                                                                   updatedExpunge.ToArray(),
                                                                   updatedExists,
                                                                   updatedRecent),
                                                "message appended"));
        }
        else if (status.Status == ImapProtocolStatus.No) {
          // 6.3.11. APPEND Command
          //       If the destination mailbox does not exist, a server MUST return an
          //       error, and MUST NOT automatically create the mailbox.  Unless it
          //       is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the APPEND
          //       if the CREATE is successful.
          bool? tryCreate;

          // RFC 2193 IMAP4 Mailbox Referrals
          // http://tools.ietf.org/html/rfc2193
          // 4.1. SELECT, EXAMINE, DELETE, SUBSCRIBE, UNSUBSCRIBE, STATUS and APPEND
          //      Referrals
          ImapUri[] referrals;

          if (ImapStatusResponseConverter.FromAppendNoResponse(status, out tryCreate, out referrals)) {
            if (tryCreate != null)
              Conclude(new ImapResponse(ImapTransactionStatus.TryCreate, status.Text));
            else if (referrals != null)
              Conclude(new ImapMailboxReferralResponse(ImapTransactionStatus.MailboxReferral, referrals, status.Text));
          }
          else {
            ConcludeError(status);
          }
        }
        else {
          ConcludeError(status);
        }
      }
    }

    private readonly List<ImapMessage> updatedMessages = new List<ImapMessage>();
    private readonly List<int> updatedExpunge = new List<int>();
    private int? updatedExists;
    private int? updatedRecent;
  }
}