using System;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal class CopyTransaction : TransactionBase {

    // ProcessArgumentNotSetted
    //   x Conclude
    // ProcessCopy
    //   ->WaitForCopy
    // WaitForCopy
    //   x Conclude

    public CopyTransaction(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Arguments.ContainsKey("mailbox name"))
        return ProcessArgumentNotSetted;
      else if (!(Request.Arguments.ContainsKey("uid set") || Request.Arguments.ContainsKey("sequence set")))
        return ProcessArgumentNotSetted;
      else
        return ProcessCopy;
    }

    private void ProcessArgumentNotSetted()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "arguments 'mailbox name' and 'uid set' or 'sequence set' must be setted"));
    }

    private void ProcessCopy()
    {
      if (Request.Arguments.ContainsKey("uid set")) {
        // UID COPY
        Send("uid copy",
             WaitForCopy,
             Request.Arguments["uid set"],
             new ImapQuotedString(ImapEncoding.ToModifiedUTF7String((string)Request.Arguments["mailbox name"])));
      }
      else {
        // COPY
        Send("copy",
             WaitForCopy,
             Request.Arguments["sequence set"],
             new ImapQuotedString(ImapEncoding.ToModifiedUTF7String((string)Request.Arguments["mailbox name"])));
      }
    }

    private void WaitForCopy()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.Tag != LastTag)
          return;

        if (status.Status == ImapProtocolStatus.Ok) {
          Conclude(new ImapResponse(ImapTransactionStatus.Ok, "message copied"));
        }
        else if (status.Status == ImapProtocolStatus.No) {
          // 6.4.7. COPY Command
          //       If the destination mailbox does not exist, a server SHOULD return
          //       an error.  It SHOULD NOT automatically create the mailbox.  Unless
          //       it is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the COPY if
          //       the CREATE is successful.
          bool? tryCreate;

          // RFC 2193 IMAP4 Mailbox Referrals
          // http://tools.ietf.org/html/rfc2193
          // 4.4. COPY Referrals
          ImapUri[] referrals;

          if (ImapStatusResponseConverter.FromCopyNoResponse(status, out tryCreate, out referrals)) {
            if (tryCreate != null)
              Conclude(new ImapResponse(ImapTransactionStatus.TryCreate, status.Text));
            else if (referrals != null)
              Conclude(new ImapMailboxReferralResponse(ImapTransactionStatus.MailboxReferral, referrals, status.Text));
          }
          else {
            ConcludeError(status);
          }
        }
        else {
          ConcludeError(status);
        }
      }
    }
  }
}
