using System;
using System.Collections.Generic;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal class IdleTransaction : TransactionBase {

    // ProcessAlreadyIdling
    //   x Conclude
    // ProcessIncalable
    //   x Conclude
    // ProcessIdle
    //   ->WaitForIdle
    //   x Conclude
    // WaitForIdle
    //   x Conclude

    public IdleTransaction(ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (Connection.IsIdling)
        return ProcessAlreadyIdling;
      else if (!Connection.ServerInfo.Capabilities.Has(ImapKnownCapabilities.Idle))
        return ProcessIncalable;
      else
        return ProcessIdle;
    }

    private void ProcessAlreadyIdling()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.Done, "already idling"));
    }

    private void ProcessIdle()
    {
      Send("idle", WaitForIdle);
    }

    private void WaitForIdle()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapCommandContinuationRequest) {
        Connection.IsIdling = true;
      }
      else if (response is ImapDataResponse) {
        ImapMessage messageData;
        int? expunge;
        int? exists;
        int? recent;
        if (ImapDataResponseConverter.FromIdleCommandResponse(response as ImapDataResponse,
                                                              out messageData,
                                                              out expunge,
                                                              out exists,
                                                              out recent)) {
          if (messageData != null)    updatedMessages.Add(messageData);
          if (expunge != null)        updatedExpunge.Add((int)expunge);
          if (exists != null)         updatedExists = exists;
          if (recent != null)         updatedRecent = recent;

          // callback
          if (Connection.IdlingStatusUpdatedCallback != null)
            Connection.IdlingStatusUpdatedCallback.BeginInvoke(new ImapSizeStatus(messageData,
                                                                                  expunge,
                                                                                  exists,
                                                                                  recent),
                                                               null,
                                                               null);        }
      }
      else if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.Tag != LastTag)
          return;

        Connection.IsIdling = false;
        Connection.UpdatedStatusWhileIdling = new ImapSizeStatus(updatedMessages.ToArray(),
                                                                 updatedExpunge.ToArray(),
                                                                 updatedExists,
                                                                 updatedRecent);

        if (status.Status == ImapProtocolStatus.Ok) {
          Conclude(new ImapResponse(ImapTransactionStatus.Ok, "idle done"));
        }
        else {
          ConcludeError(status);
        }
      }
    }

    private readonly List<ImapMessage> updatedMessages = new List<ImapMessage>();
    private readonly List<int> updatedExpunge = new List<int>();
    private int? updatedExists;
    private int? updatedRecent;
  }
}