using System;
using System.Threading;

using Smdn.Protocols.Pop3.Transaction.Transactions;
using Smdn.Protocols.Pop3.Protocol;

namespace Smdn.Protocols.Pop3.Transaction {
  // Layers:
  //   Client.PopClient
  //   Client.PopClientAgent
  //   Session.PopSessionAgent
  //   Session.PopSession
  //   Transaction.PopTransactionAgent
  // * Transaction.PopTransaction/Transaction.PopRequest/Transaction.PopResponse
  //   Protocol.PopConnection/Protocol.PopProtocolCommand/Protocol.PopProtocolResponse
  //   (POP3)

  public abstract class PopTransaction {
    protected delegate void ProcessTransactionDelegate();

    public PopConnection Connection {
      get { return connection; }
    }

    public PopRequest Request {
      get { return request; }
    }

    public PopResponse Response {
      get { return response; }
      protected internal set { response = value; }
    }

    internal int SequenceId {
      get { return sequenceId; }
    }

    protected PopTransaction(PopConnection connection, PopRequest request)
    {
      this.connection = connection;
      this.request = request;
      this.response = null;
    }

    public override string ToString()
    {
      return string.Format("{{Connection={0}, Request={1}, Response={2}}}", connection, request, response);
    }

    public static PopTransaction Create(PopConnection connection, PopRequest request)
    {
      var sequenceId = Interlocked.Increment(ref sequenceIdCount);

      if (sequenceIdCount == int.MaxValue)
        Interlocked.Exchange(ref sequenceIdCount, 0);

      PopTransaction transaction;

      switch (request.Method) {
        /*
         * transaction for initial server greeting
         */
        case PopTransactionMethod.Greeting:
          transaction = new GreetingTransaction(connection, request);
          break;

        /*
         * transactions in any state
         */
        case PopTransactionMethod.Quit:
          transaction = new QuitTransaction(connection, request);
          break;

        case PopTransactionMethod.Capa:
          transaction = new CapaTransaction(connection, request);
          break;

        /*
         * transactions in authrozation state
         */
        case PopTransactionMethod.UserPass:
          transaction = new CompositeTransaction(connection, request, new [] {
            new CompositeTransaction.InnerTransaction(new UserTransaction(connection, request), true),
            new CompositeTransaction.InnerTransaction(new PassTransaction(connection, request), true),
          });
          break;

        case PopTransactionMethod.Apop:
          transaction = new ApopTransaction(connection, request);
          break;

        case PopTransactionMethod.Auth:
          transaction = new AuthTransaction(connection, request);
          break;

        case PopTransactionMethod.Stls:
          transaction = new StlsTransaction(connection, request);
          break;

        /*
         * transactions in transaction state
         */

        case PopTransactionMethod.Stat:
          transaction = new StatTransaction(connection, request);
          break;

        case PopTransactionMethod.List:
          transaction = new ListTransaction(connection, request);
          break;

        case PopTransactionMethod.Retr:
          transaction = new RetrTransaction(connection, request);
          break;

        case PopTransactionMethod.Dele:
          transaction = new DeleTransaction(connection, request);
          break;

        case PopTransactionMethod.NoOp:
          transaction = new NoOpTransaction(connection, request);
          break;

        case PopTransactionMethod.Rset:
          transaction = new RsetTransaction(connection, request);
          break;

        case PopTransactionMethod.Top:
          transaction = new TopTransaction(connection, request);
          break;

        case PopTransactionMethod.Uidl:
          transaction = new UidlTransaction(connection, request);
          break;

        default:
          throw new InvalidOperationException("invalid request");
      }

      transaction.ResetInternal();
      transaction.sequenceId = sequenceId;

      return transaction;
    }

    public virtual void Abort()
    {
      // TODO: 中断前に受信してキューに残っているレスポンスの考慮
      lock (this) {
        Transit(null);

        response = new PopResponse(PopTransactionStatus.RequestTimeout, "request timed out");
      }
    }

    protected abstract ProcessTransactionDelegate Reset();

    internal void ResetInternal()
    {
      currentProcess = Reset();
    }

    public void Process()
    {
      // トランザクションを処理する
      try {
        if (currentProcess != null)
          currentProcess();
      }
      catch (Exception ex) {
        Trace.Log(ex);

        response = new PopResponse(PopTransactionStatus.BadRequest, ex.Message);

        Transit(null);
      }
    }

    protected void Transit(ProcessTransactionDelegate process)
    {
      currentProcess = process;
    }

    private static int sequenceIdCount = 0;
    private readonly PopConnection connection;
    private readonly PopRequest request;
    private PopResponse response;
    private int sequenceId;
    private ProcessTransactionDelegate currentProcess;
  }
}
