using System;
using System.Text;
using System.Security.Cryptography;

using Smdn.Protocols.Pop3.Protocol;

namespace Smdn.Protocols.Pop3.Transaction.Transactions {
  internal class StlsTransaction : TransactionBase {

    // ProcessStlsIncapable
    //   x Conclude
    // ProcessBadRequest
    //   x Conclude
    // ProcessStls
    //   ->WaitForStls
    // WaitForStls
    //   x Conclude

    public StlsTransaction(PopConnection connection, PopRequest request)
      : base(connection, request)
    {
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Connection.ServerInfo.Capabilities.Has(PopKnownCapabilities.Stls))
        return ProcessStlsIncapable;
      else if (!(Request is PopStlsRequest))
        return ProcessBadRequest;
      else
        return ProcessStls;
    }

    private void ProcessStlsIncapable()
    {
      Conclude(new PopResponse(PopTransactionStatus.Incapable, "STLS incapable"));
    }

    private void ProcessBadRequest()
    {
      Conclude(new PopResponse(PopTransactionStatus.BadRequest, "bad request"));
    }

    private void ProcessStls()
    {
      Send("STLS");

      Transit(WaitForStls);
    }

    private void WaitForStls()
    {
      PopProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is PopStatusResponse) {
        var status = response as PopStatusResponse;

        if (status.Status == PopProtocolStatus.Positive) {
          // RFC 2595 Using TLS with IMAP, POP3 and ACAP
          // http://tools.ietf.org/html/rfc2595
          // 4. POP3 STARTTLS extension
          //              A TLS negotiation begins immediately after the CRLF at the
          //              end of the +OK response from the server.  A -ERR response
          //              MAY result if a security layer is already active.  Once a
          //              client issues a STLS command, it MUST NOT issue further
          //              commands until a server response is seen and the TLS
          //              negotiation is complete.
          Connection.UpgradeSecurity(Connection.ServerInfo.Uri,
                                     (Request as PopStlsRequest).CreateAuthenticatedStreamCallback);

          //              Once TLS has been started, the client MUST discard cached
          //              information about server capabilities and SHOULD re-issue
          //              the CAPA command.  This is necessary to protect against
          //              man-in-the-middle attacks which alter the capabilities list
          //              prior to STLS.  The server MAY advertise different
          //              capabilities after STLS.
          Connection.ServerInfo.Capabilities = new PopCapability();

          Conclude(new PopResponse(PopTransactionStatus.Ok, "succeeded"));
        }
        else {
          ConcludeError(status);
        }
      }
    }
  }
}
